@@ -3,5 +3,6 @@ from . import geometry
 from . import dynamics
 from . import solvers
 from . import draw
+from . import stores
 
-from .json_parser import load_spec, save_spec, load_default
+from .json_parser import load_spec, save_spec
