@@ -4,6 +4,12 @@ import os
 CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
 
 
+def quasistatic_bulk_spec():
+    specfile = os.path.join(CURRENT_DIR, 'bulk_qs.json')
+    return load_spec(specfile)
+
+
+
 def quasistatic_plane_spec():
     """Default specification for the dynamics
     of a sheet vertex model - also suitable for general faceted tissue
@@ -22,6 +28,7 @@ def quasistatic_plane_spec():
              "radial_tension": 0.0
              },
         "edge": {
+             "is_active": 1,
              "line_tension": 0.12
              },
         "settings": {
