@@ -1,16 +1,3 @@
-import numpy as np
-import pandas as pd
-
-# from .. import libtyssue_core as libcore
-# from . import generation
-from ..utils.utils import set_data_columns, spec_updater
-from ..config.json_parser import load_default
-
-import warnings
-
-import logging
-log = logging.getLogger(name=__name__)
-
 '''
 Core definitions
 
@@ -60,12 +47,18 @@ default values is allways infered from the python parsed type. Thus
         }
 '''
 
-#- BC -#
-# Classes must inherit from Object in order for setters
-# decorators to work properly according to some
-# stackoverflow post, but it seems to work properly as
-# it is right now.
-class Epithelium():
+
+import warnings
+import logging
+import numpy as np
+import pandas as pd
+from collections import deque
+
+from ..utils.utils import set_data_columns, spec_updater
+log = logging.getLogger(name=__name__)
+
+
+class Epithelium:
     '''
     The whole tissue.
 
@@ -76,59 +69,66 @@ class Epithelium():
         '''
         Creates an epithelium
 
-        Parameters:
-        -----------
-        identifier: string
-        datasets: dictionary of dataframes
-        the datasets dict specifies the names, data columns
-        and value types of the modeled tyssue
+        Parameters
+        ----------
+        identifier : string
+        datasets : dictionary of dataframes
+          the datasets dict specifies the names, data columns
+          and value types of the modeled tyssue
+
+        Note
+        ----
+        For efficiency reasons, we have to maintain monotonous RangeIndex
+        for each dataset. Thus, _the index of an element can change_,
+        and should not be used as an identifier.
 
         '''
-        if coords is None:
-            coords = ['x', 'y', 'z']
-        self.coords = coords
-        # edge's dx, dy, dz
-        self.dcoords = ['d'+c for c in self.coords]
-        self.dim = len(self.coords)
-        # edge's normals
-        if self.dim == 3:
-            self.ncoords = ['n'+c for c in self.coords]
+        # backup container
+        # TODO: pass the max backup number as a config argument
+        self._backups = deque(maxlen=5)
 
         # each of those has a separate dataframe, as well as entries in
         # the specification files
-        frame_types = {'edge', 'vert', 'face',
-                       'cell'}
-
-        # # Really just to ensure the debugger is silent
-        # [self.edge_df,
-        #  self.vert_df,
-        #  self.face_df,
-        #  self.cell_df] = [None, ] * 4
-
+        _frame_types = {'edge', 'vert', 'face', 'cell'}
         self.identifier = identifier
-        if not set(datasets).issubset(frame_types):
+        if not set(datasets).issubset(_frame_types):
             raise ValueError('The `datasets` dictionnary should'
-                             ' contain keys in {}'.format(frame_types))
+                             ' contain keys in {}'.format(_frame_types))
         self.datasets = datasets
-        # for name, data in datasets.items():
-        #     setattr(self, '{}_df'.format(name), data)
         self.data_names = list(datasets.keys())
         self.element_names = ['srce', 'trgt',
                               'face', 'cell'][:len(self.data_names)]
+        # Infer specs from the first rows of the datasets
         if specs is None:
-            specs = {name: {} for name in self.data_names}
+            specs = {elem : df.iloc[0].to_dict() for elem, df in datasets.items()}
         if 'settings' not in specs:
             specs['settings'] = {}
 
         self.specs = specs
         self.update_specs(specs, reset=False)
-        self.edge_mindex = pd.MultiIndex.from_arrays(self.edge_idx.values.T,
-                                                     names=self.element_names)
-        # # Topology (geometry independant)
-        self.reset_topo()
+
+        if coords is None:
+            coords = [c for c in 'xyz' if c in datasets['vert'].columns]
+
+        self.coords = coords
+        # edge's dx, dy, dz
+        self.dcoords = ['d'+c for c in self.coords]
+        # edge's unit length vector
+        self.ucoords = ['u'+c for c in self.coords]
+
+        self.dim = len(self.coords)
+        # edge's normals
+        if self.dim == 3:
+            self.ncoords = ['n'+c for c in self.coords]
+
         self.bbox = None
         self.set_bbox()
 
+    @property
+    def edge_mindex(self):
+        return pd.MultiIndex.from_arrays(self.edge_idx.values.T,
+                                         names=self.element_names)
+
     @property
     def face_df(self):
         return self.datasets['face']
@@ -147,12 +147,7 @@ class Epithelium():
 
     @property
     def cell_df(self):
-        #- BC -#
-        # Recommends to test if the epithelium has a 'cell' df before returning,
-        # some epitheliums don't require this object (2D)
-        # Should it raise a warning/error if the user asks for
-        # a cell_df from an eptm that doesn't have one ?
-        return self.datasets['cell']
+        return self.datasets.get('cell', None)
 
     @cell_df.setter
     def cell_df(self, value):
@@ -166,20 +161,6 @@ class Epithelium():
     def vert_df(self, value):
         self.datasets['vert'] = value
 
-    # @property
-    # def datasets(self):
-    #     datasets = {element: getattr(self, '{}_df'.format(element))
-    #                 for element in self.data_names}
-    #     return datasets
-
-    # # @datasets.getter
-    # # def datasets(self, level):
-    # #     return getattr(self, '{}_df'.format(level))
-
-    # @datasets.setter
-    # def datasets(self, level, new_df):
-    #     setattr(self, '{}_df'.format(level), new_df)
-
     def copy(self, deep_copy=True):
         """
         Returns a copy of the epithelium
@@ -201,59 +182,53 @@ class Epithelium():
             datasets = self.datasets
 
         identifier = self.identifier+'_copy'
-        new = Epithelium(identifier, datasets,
-                         specs=self.specs, coords=None)
+        new = type(self)(identifier, datasets,
+                         specs=self.specs, coords=self.coords)
         return new
 
+    def backup(self):
+        """Creates a copy of self and keeps a reference to it
+        in the self._backups deque.
+
+        """
+        log.info('Backing up')
+        self._backups.append(self.copy(deep_copy=True))
+
+    def restore(self):
+        '''Resets the eptithelium data to its last backed up state
+
+        A copy of the current state prior to restoring is kept in the
+        `_bad` attribute for inspection.
+        '''
+
+        log.info('Restoring')
+        log.info('a copy of the unrestored epithelium'
+                 ' is stored in the `_bad` attribute')
+        bck = self._backups.pop()
+        self._bad = self.copy(deep_copy=True)
+        self.datasets = bck.datasets
+        self.specs = bck.specs
+
     @property
     def settings(self):
         return self.specs['settings']
 
-    @settings.setter
-    def settings(self, key, value): #pragma: no cover
-        # (Actually the 'settings' getter is called
-        # and then the dictionary class setter
-        # instead of directly the 'settings' setter.)
-        # See http://stackoverflow.com/a/3137768
-        # So this method is not actually called.
-        self.specs['settings'][key] = value
-
     def update_specs(self, new, reset=False):
 
         spec_updater(self.specs, new)
         set_data_columns(self.datasets, new, reset)
 
-    def set_specs(self, domain, base,
-                  new_specs=None,
-                  default_base=None, reset=False):  # pragma: no cover
-        warnings.warn('Deprecated, use update_specs() instead.')
-
-        if base is None:
-            self.update_specs(load_default(domain, default_base), reset)
-        else:
-            self.update_specs(load_default(domain, base), reset)
-        if new_specs is not None:
-            self.update_specs(new_specs, reset)
-
     def update_num_sides(self):
-        self.face_df['num_sides'] = self.edge_df.face.value_counts().loc[
-            self.face_df.index]
+        self.face_df['num_sides'] = self.edge_df.face.value_counts()
 
     def update_num_faces(self):
         self.cell_df['num_faces'] = self.edge_df.groupby('cell').apply(
             lambda df: df['face'].unique().size)
 
-    def update_mindex(self):
-        self.edge_mindex = pd.MultiIndex.from_arrays(self.edge_idx.values.T,
-                                                     names=self.element_names)
-
     def reset_topo(self):
         self.update_num_sides()
-        self.update_mindex()
         if 'cell' in self.data_names:
             self.update_num_faces()
-        if 'opposite' in self.edge_df.columns:
-            self.edge_df['opposite'] = get_opposite(self.edge_df)
 
     @property
     def face_idx(self):
@@ -278,6 +253,7 @@ class Epithelium():
             return self.cell_df.shape[0]
         elif 'face' in self.data_names:
             return self.face_df.shape[0]
+        return None
 
     @property
     def Nv(self):
@@ -314,8 +290,8 @@ class Epithelium():
                           self.e_face_idx)).T
 
     def _upcast(self, idx, df):
-
-        upcast = df.loc[idx]
+        ## Assumes a flat index
+        upcast = df.take(idx)
         upcast.index = self.edge_df.index
         return upcast
 
@@ -365,9 +341,16 @@ class Epithelium():
         return self._upcast(self.edge_df['cell'], df)
 
     def _lvl_sum(self, df, lvl):
-        df_ = df.copy()
-        df_.index = self.edge_mindex
-        return df_.sum(level=lvl)
+        df_ = df
+        if isinstance(df, np.ndarray):
+            df_ = pd.DataFrame(df, index=self.edge_df.index)
+            df_[lvl] = self.edge_df[lvl]
+        elif isinstance(df, pd.Series):
+            df_ = df.to_frame()
+        elif lvl not in df.columns:
+            df_ = df.copy()
+        df_[lvl] = self.edge_df[lvl]
+        return df_.groupby(lvl).sum()
 
     def sum_srce(self, df):
         return self._lvl_sum(df, 'srce')
@@ -410,140 +393,56 @@ class Epithelium():
             lambda df: df[periph])
         return orbits
 
+    def idx_lookup(self, elem_id, element):
+        """returns the current index of the element
+        with the 'id' column equal to elem_id
+        """
+        df = self.datasets[element]['id']
+        idx = df[df==elem_id].index
+        if len(idx):
+            return idx[0]
+        else:
+            return None
+
     def face_polygons(self, coords):
         def _get_verts_pos(face):
-            try:
-                edges = _ordered_edges(face)
-            except IndexError:
-                #- BC -#
-                # I'm still trying to figure
-                # out a way to raise this exception
-                # with altered datasets but to no avail
-                # Leaving it included in coverage.
-                log.warning('Face is not closed')
-                return np.nan
+            edges = _ordered_edges(face)
             return np.array([self.vert_df.loc[idx[0], coords]
                              for idx in edges])
         polys = self.edge_df.groupby('face').apply(_get_verts_pos).dropna()
         return polys
 
-    def get_extra_indices(self):
-        """Computes extra indices:
+    def validate(self):
+        """returns True if the mesh is validated
 
-        - `self.free_edges`: half-edges at the epithelium boundary
-        - `self.dble_edges`: half-edges inside the epithelium,
-          with an opposite
-        - `self.east_edges`: half of the `dble_edges`, pointing east
-          (figuratively)
-        - `self.west_edges`: half of the `dble_edges`, pointing west
-           (the order of the east and west edges is conserved, so that
-           the ith west half-edge is the opposite of the ith east half-edge)
-        - `self.sgle_edges`: joint index over free and east edges, spanning
-           the entire graph without double edges
-        - `self.wrpd_edges`: joint index over free edges followed by the
-           east edges twice, such that a vector over the whole half-edge
-            dataframe is wrapped over the single edges
-        - `self.srtd_edges`: index over the whole half-edge sorted such that
-           the free edges come first, then the east, then the west
-
-        Also computes:
-        - `self.Ni`: the number of inside full edges
-          (i.e. `len(self.east_edges)`)
-        - `self.No`: the number of outside full edges
-          (i.e. `len(self.free_edges)`)
-        - `self.Nd`: the number of double half edges
-          (i.e. `len(self.dble_edges)`)
-        - `self.anti_sym`: `pd.Series` with shape `(self.Ne,)`
-          with 1 at the free and east half-edges and -1
-          at the opposite half-edges.
-
-        Notes
-        -----
-
-        - East and west is resepctive to some orientation at the
-          moment the indices are computed the partition stays valid as
-          long as there are no changes in the topology, so due to vertex
-          displacement, 'east' and 'west' might not stay valid. This is
-          just a practical naming convention.
-
-        - As the name suggest, this method is not working for edges in
-          3D pointing *exactly* north or south, ie iff `edge['dx'] ==
-          edge['dy'] == 0`. Until we need or find a better solution,
-          we'll just assert it worked.
+        e.g. has only closed polygons and polyhedra
         """
-
-        if 'opposite' not in self.edge_df.columns:
-            self.edge_df['opposite'] = get_opposite(self.edge_df)
-
-        self.dble_edges = self.edge_df[self.edge_df['opposite'] >= 0].index
-        theta = np.arctan2(self.edge_df.loc[self.dble_edges, 'dy'],
-                           self.edge_df.loc[self.dble_edges, 'dx'])
-
-        self.east_edges = self.edge_df.loc[self.dble_edges][
-            (theta >= 0) & (theta < np.pi)].index
-        self.west_edges = pd.Index(self.edge_df.loc[
-            self.east_edges, 'opposite'].astype(np.int), name='edge')
-
-        self.free_edges = self.edge_df[self.edge_df['opposite'] == -1].index
-        self.sgle_edges = self.free_edges.append(self.east_edges)
-        self.srtd_edges = self.sgle_edges.append(self.west_edges)
-
-        # Index over the east and free edges, then the opposite indexed
-        # by their east counterpart
-        self.wrpd_edges = self.sgle_edges.append(self.east_edges)
-
-        self.Ni = self.east_edges.size  # number of inside (east) edges
-        self.Nd = self.dble_edges.size  # number of non free half edges
-        self.No = self.free_edges.size  # number of free halfedges
-        try:
-            assert (2*self.Ni + self.No) == self.Ne
-            assert self.west_edges.size == self.Ni
-            assert self.Nd == 2*self.Ni
-        #- BC -#
-        # Not sure how to build
-        # input data so the partition
-        # fails (so we can see
-        # if the exception is
-        # correctly raised).
-        # Leaving it in the coverage
-        # anyway.
-        except AssertionError:
-            raise AssertionError('''
-            Inconsistent partition:
-            total half-edges: %s
-            number of free: %s
-            number of east: %s
-            number of west: %s''' % (self.Ne, self.No, self.Ni,
-                                     self.west_edges.size))
-
-        # Anti symetric vector (1 at east and free edges, -1 at opposite)
-        self.anti_sym = pd.Series(np.ones(self.Ne),
-                                  index=self.edge_df.index)
-        self.anti_sym.loc[self.west_edges] = -1
-
-    def sort_edges_eastwest(self):
-        """reorder edges such the free edges are first,
-        then the first half of the double edges, then the other half of
-        the double edges, this way, each subset of the edges dataframe
-        are contiguous.
-        """
-        self.get_extra_indices()
-        self.edge_df = self.edge_df.loc[self.srtd_edges]
-        self.reset_index()
-        self.reset_topo()
-        self.get_extra_indices()
+        return np.alltrue(~self.get_invalid())
 
     def get_valid(self):
-        """Set true if the face is a closed polygon
+        """Set the 'is_valid' column to true if the faces are all closed polygons,
+        and the cells closed polyhedra.
         """
-        is_valid = self.edge_df.groupby('face').apply(_test_valid)
-        self.edge_df['is_valid'] = self.upcast_face(is_valid)
+        is_valid_face = self.edge_df.groupby('face').apply(_test_valid)
+        is_valid = self.upcast_face(is_valid_face)
+        if 'cell' in self.data_names:
+            is_valid_cell = self.edge_df.groupby('cell').apply(
+                _is_closed_cell)
+            is_valid = is_valid | self.upcast_cell(is_valid_cell)
+        self.edge_df['is_valid'] = is_valid
+        return is_valid
 
     def get_invalid(self):
-        """Returns a mask over edge for invalid faces
+        """Returns a mask over self.edge_df for invalid faces
         """
-        is_invalid = self.edge_df.groupby('face').apply(_test_invalid)
-        return self.upcast_face(is_invalid)
+        is_invalid_face = self.edge_df.groupby('face').apply(_test_invalid)
+        invalid_edges = self.upcast_face(is_invalid_face)
+        if 'cell' in self.data_names:
+            is_invalid_cell = 1 - self.edge_df.groupby('cell').apply(
+                _is_closed_cell)
+            invalid_edges = invalid_edges | self.upcast_cell(is_invalid_cell)
+        self.edge_df['is_valid'] = ~invalid_edges
+        return invalid_edges
 
     def sanitize(self):
         """Removes invalid faces and associated vertices
@@ -556,11 +455,14 @@ class Epithelium():
         the cells and faces containing those edges
         """
         top_level = self.element_names[-1]
-        log.info('Removing cells at the {} level'.format(top_level))
+        log.info('Removing cells at the %s level', top_level)
         fto_rm = self.edge_df.loc[edge_out, top_level].unique()
-        if not len(fto_rm):
+        if not fto_rm.shape[0]:
             log.info('Nothing to remove')
             return
+        if fto_rm.shape[0] == self.datasets[top_level].shape[0]:
+            raise ValueError('sanitize would delete the whole epithlium')
+
         fto_rm.sort()
         log.info('{} {} level elements will be removed'.format(len(fto_rm),
                                                                top_level))
@@ -622,11 +524,15 @@ class Epithelium():
 
         new_vertidx = pd.Series(np.arange(self.vert_df.shape[0]),
                                 index=self.vert_df.index)
-        self.edge_df['srce'] = self.upcast_srce(new_vertidx)
-        self.edge_df['trgt'] = self.upcast_trgt(new_vertidx)
+        # Here we use loc and not the take from upcast
+
+        self.edge_df['srce'] = new_vertidx.loc[self.edge_df['srce']].values
+        self.edge_df['trgt'] = new_vertidx.loc[self.edge_df['trgt']].values
+
         new_fidx = pd.Series(np.arange(self.face_df.shape[0]),
                              index=self.face_df.index)
-        self.edge_df['face'] = self.upcast_face(new_fidx)
+
+        self.edge_df['face'] = new_fidx.loc[self.edge_df['face']].values
 
         self.vert_df.reset_index(drop=True, inplace=True)
         self.vert_df.index.name = 'vert'
@@ -637,7 +543,7 @@ class Epithelium():
         if 'cell' in self.data_names:
             new_cidx = pd.Series(np.arange(self.cell_df.shape[0]),
                                  index=self.cell_df.index)
-            self.edge_df['cell'] = self.upcast_cell(new_cidx)
+            self.edge_df['cell'] = new_cidx.loc[self.edge_df['cell']].values
             self.cell_df.reset_index(drop=True, inplace=True)
             self.cell_df.index.name = 'cell'
 
@@ -703,7 +609,7 @@ class Epithelium():
 
 def _ordered_edges(face_edges):
     """Returns "srce", "trgt" and "face" indices
-    organized clockwise for each edge.
+    organized clockwise for each face.
 
     Parameters
     ----------
@@ -715,7 +621,8 @@ def _ordered_edges(face_edges):
     edges: list of 3 ints
         srce, trgt, face indices, ordered
     """
-    srces, trgts, faces = face_edges[['srce', 'trgt', 'face']].values.T
+
+    srces, trgts, faces = (face_edges[col].values for col in ['srce', 'trgt', 'face'])
     srce, trgt, face_ = srces[0], trgts[0], faces[0]
     edges = [[srce, trgt, face_]]
     for face_ in faces[1:]:
@@ -748,18 +655,63 @@ def _test_valid(face):
     return s1 == s2
 
 
-def get_opposite(edge_df):
-    """
-    Returns the indices opposite to the edges in `edge_df`
-    """
-    st_indexed = edge_df[['srce', 'trgt']].reset_index().set_index(
-        ['srce', 'trgt'], drop=False)
-    flipped = st_indexed.index.swaplevel(0, 1)
-    flipped.names = ['srce', 'trgt']
-    opposite = st_indexed.loc[flipped, 'edge'].values
-    opposite[np.isnan(opposite)] = -1
+def get_opposite_faces(eptm):
+    """Populates the 'opposite' column of eptm.face_df with the index of
+    the opposite face or -1 if the face has no opposite.
 
-    return opposite.astype(np.int)
+    """
+    face_v = eptm.edge_df.groupby('face').apply(lambda df: frozenset(df['srce']))
+    face_v2 = pd.Series(data=face_v.index, index=face_v.values)
+    grouped = face_v2.groupby(level=0)
+    cardinal = grouped.apply(len)
+    if cardinal.max() > 2:
+        raise ValueError('Invalid topology,'
+                         ' incorrect faces: {}'.format(cardinal.argmax()))
+    eptm.face_df['opposite'] = -1
+
+    face_pairs = []
+    grouped.apply(lambda s: face_pairs.append(list(s.values))
+                  if len(s) == 2 else np.nan).dropna()
+    face_pairs = np.array(face_pairs)
+    eptm.face_df.loc[face_pairs[:, 0], 'opposite'] = face_pairs[:, 1]
+    eptm.face_df.loc[face_pairs[:, 1], 'opposite'] = face_pairs[:, 0]
+
+
+def _next_edge(edf):
+
+    edf['edge'] = edf.index
+    next_edge = edf.set_index(
+        'srce', append=False).loc[
+            edf['trgt'], 'edge'].values
+    return pd.Series(index=edf.index, data=next_edge)
+
+def _prev_edge(edf):
+
+    edf['edge'] = edf.index
+    next_edge = edf.set_index(
+        'trgt', append=False).loc[
+            edf['srce'], 'edge'].values
+    return pd.Series(index=edf.index, data=next_edge)
+
+
+def get_next_edges(sheet):
+    '''
+    returns a pd.Series with the index of the next
+    edge for each edge
+    '''
+    next_e = sheet.edge_df.groupby('face').apply(_next_edge)
+    next_e.index = next_e.index.droplevel('face')
+    return next_e.sort_index()
+
+
+def get_prev_edges(sheet):
+    '''
+    returns a pd.Series with the index of the next
+    edge for each edge
+    '''
+    prev_e = sheet.edge_df.groupby('face').apply(_prev_edge)
+    prev_e.index = prev_e.index.droplevel('face')
+    return prev_e.sort_index()
 
 
 def _is_closed_cell(e_df):
